from json import loads
from pathlib import Path
from enum import Enum


_cidades = None


def ler_cidades():
    global _cidades
    if _cidades is not None:
        return _cidades

    with open(Path(__file__).parent / "municipios.json", encoding="utf-8") as f:
        _cidades = loads(f.read())
    return _cidades


def validar_codigo_cidade(codigo: int) -> bool:
    cidades = ler_cidades()
    return any(cidade["id"] == codigo for cidade in cidades)


def normalizar_uf(uf: str) -> str:
    cidades = ler_cidades()
    uf = uf.upper()
    assert any(
        cidade["regiao-imediata"]["regiao-intermediaria"]["UF"]["sigla"] == uf
        for cidade in cidades
    ), "UF não encontrada. Insira uma UF válida no formato de sigla (ex: SP, RJ, MG, etc)"
    return uf


class UF(Enum):
    ACRE = "AC"
    ALAGOAS = "AL"
    AMAPA = "AP"
    AMAZONAS = "AM"
    BAHIA = "BA"
    CEARA = "CE"
    DISTRITO_FEDERAL = "DF"
    ESPIRITO_SANTO = "ES"
    GOIAS = "GO"
    MARANHAO = "MA"
    MATO_GROSSO = "MT"
    MATO_GROSSO_DO_SUL = "MS"
    MINAS_GERAIS = "MG"
    PARA = "PA"
    PARAIBA = "PB"
    PARANA = "PR"
    PERNAMBUCO = "PE"
    PIAUI = "PI"
    RIO_DE_JANEIRO = "RJ"
    RIO_GRANDE_DO_NORTE = "RN"
    RIO_GRANDE_DO_SUL = "RS"
    RONDONIA = "RO"
    RORAIMA = "RR"
    SANTA_CATARINA = "SC"
    SAO_PAULO = "SP"
    SERGIPE = "SE"
    TOCANTINS = "TO"



from lxml.etree import ElementBase
from typing import List


def assert_xml_iguais(
    xml1: ElementBase,
    xml2: ElementBase,
    path: List[str] = [],
    ignorar_tags: List[str] = [],
):
    xml1: ElementBase = xml1
    xml2: ElementBase = xml2

    path_str = " -> ".join(path)

    # mesma tag
    if xml1.tag in ignorar_tags and xml2.tag in ignorar_tags:
        return

    assert xml1.tag == xml2.tag, f"Tags diferentes: {path_str} {xml1.tag} != {xml2.tag}"

    # mesmos atributos
    for k, v in xml1.items():
        assert (
            xml2.get(k) == v
        ), f"Atributos diferentes: {path_str} {k} {v} != {xml2.get(k)}"

    # mesmos filhos
    children1 = [c1.tag for c1 in xml1 if c1.tag not in ignorar_tags]
    children2 = [c2.tag for c2 in xml2 if c2.tag not in ignorar_tags]
    assert (
        children1 == children2
    ), f"Filhos diferentes: {[c1.tag for c1 in xml1]} != {[c2.tag for c2 in xml2]}"

    if len(xml1) == 0:
        assert (
            xml1.text == xml2.text
        ), f"Textos diferentes: {path_str} -> {xml1.tag} {xml1.text} != {xml2.text}"
    else:
        for c1, c2 in zip(xml1, xml2):
            assert_xml_iguais(c1, c2, path + [xml1.tag], ignorar_tags=ignorar_tags)



def normalizar_validar_telefone(telefone: str) -> str:
    """
    Normaliza o número de telefone removendo caracteres não numéricos.
    
    Args:
        telefone (str): Número de telefone a ser normalizado.
        
    Returns:
        str: Número de telefone normalizado.
    """
    digitos = "".join(t for t in telefone if t.isdigit())
    if len(digitos) <= 11:
        return digitos
    else:
        raise ValueError(f"Número de telefone inválido '{telefone}'. Deve conter no máximo 11 dígitos. Verifique se não está adicionando DDI.")
    



from abc import abstractmethod, ABC
from abstra_notas.assinatura import Assinador
from .templates import load_template
from zeep.plugins import HistoryPlugin
from zeep import Client, Transport, Settings
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager
from lxml.etree import tostring, fromstring, ElementBase
from pathlib import Path
from typing import Generic, TypeVar
from tempfile import mktemp
import ssl



T = TypeVar('T', bound='Envio')

# Adapta para o TLS antigo de forma a ser compatível com o servidor da prefeitura do Rio de Janeiro
class TLSAdapterCorrigido(HTTPAdapter):
    """Adaptador TLS corrigido para resolver problemas de SSL"""
    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        ctx = ssl.create_default_context()
        ctx.set_ciphers("HIGH:!DH:!aNULL")
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        ctx.minimum_version = ssl.TLSVersion.TLSv1_2
        ctx.maximum_version = ssl.TLSVersion.TLSv1_3
        
        try:
            ctx.options |= ssl.OP_NO_COMPRESSION
        except AttributeError:
            pass
        
        self.poolmanager = PoolManager(
            num_pools=connections, 
            maxsize=maxsize, 
            block=block, 
            ssl_context=ctx
        )

class Envio(ABC, Generic[T]):
    def gerar_xml(self) -> ElementBase:

        xml = load_template(self.__class__.__name__).render(self.__dict__)
        return fromstring(xml.encode("utf-8"))
    
    @abstractmethod
    def nome_operacao(self):
        raise NotImplementedError("Subclasses must implement this method.")
    
    @abstractmethod
    def resposta(self, xml: ElementBase):
        """
        Método para processar a resposta do serviço e retornar um objeto de resposta.
        Deve ser implementado por subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method.")
    
    @property
    def schema_path(self) -> Path:
        return Path(__file__).parent / "schemas" / f"{self.nome_operacao()}.xsd"


    def executar(self, caminho_pfx: Path, senha_pfx: str, homologacao=False) -> T:
        assinador = Assinador(caminho_pfx, senha_pfx)
        try:
            history = HistoryPlugin()
            keyfile = Path(mktemp())
            keyfile.write_bytes(assinador.private_key_pem_bytes)
            certfile = Path(mktemp())
            certfile.write_bytes(assinador.cert_pem_bytes)

            xml = self.gerar_xml()
            session = Session()
            session.cert = (certfile, keyfile)
            session.verify = False

            session.mount('https://', TLSAdapterCorrigido())

            settings = Settings(strict=True, xml_huge_tree=True)
            transport = Transport(session=session, cache=None, timeout=500)

            if homologacao:
                url = "https://notacariocahom.rio.gov.br/WSNacional/nfse.asmx?wsdl"
            else:
                url= "https://notacarioca.rio.gov.br/WSNacional/nfse.asmx?wsdl"
            client = Client(
                url, transport=transport, settings=settings, plugins=[history]
            )

            # Assina o XML
            xml_assinado = xml
            #xml_assinado = assinador.assinar_xml(xml)

            request_tmp_path = Path(mktemp())
            request_tmp_path.write_text(tostring(xml_assinado, encoding=str), encoding="utf-8")
            #print(f"Request saved to: {request_tmp_path}")


            response: str = getattr(client.service, self.nome_operacao())( tostring(xml_assinado, encoding=str))

            response_temp_path = Path(mktemp())
            response_temp_path.write_text(response, encoding="utf-8")
            #print(f"Response saved to: {response_temp_path}")
            xml_resposta =  fromstring(response.encode("utf-8"))
            return self.resposta(xml_resposta)
        finally:
            if keyfile.exists():
                keyfile.unlink()
            if certfile.exists():
                certfile.unlink()


from unittest import TestCase
from .cancelamento_nfe import CancelamentoNFe, RetornoCancelamentoNFe
from pathlib import Path
from lxml.etree import fromstring
from .cliente import ClienteMock
from abstra_notas.assinatura import AssinadorMock
from abstra_notas.validacoes.xml_iguais import assert_xml_iguais


class CancelamentoTest(TestCase):
    def test_exemplo(self):
        assinador = AssinadorMock()
        self.maxDiff = None
        exemplo_path = Path(__file__).parent / "exemplos" / "PedidoCancelamentoNFe.xml"
        exemplo_xml = assinador.assinar_xml(fromstring(exemplo_path.read_bytes()))

        pedido = CancelamentoNFe(
            inscricao_prestador="39616924",
            numero_nfe="17945",
            remetente="99-99/999.70-00//100",
            transacao=True,
        )

        pedido_xml = assinador.assinar_xml(pedido.gerar_xml(assinador=assinador))
        assert_xml_iguais(
            pedido_xml,
            exemplo_xml,
            ignorar_tags=["AssinaturaCancelamento", "Signature"],
        )

        cliente = ClienteMock()
        resultado = cliente.cancelar_nota(pedido)

        self.assertEqual(
            resultado,
            RetornoCancelamentoNFe(),
        )



from unittest import TestCase
from .consulta import ConsultaNFePeriodo
from .cliente import ClienteMock
from abstra_notas.assinatura import AssinadorMock
from datetime import date


class ConsultaTest(TestCase):
    def test_recebidas(self):
        assinador = AssinadorMock()
        self.maxDiff = None

        pedido = ConsultaNFePeriodo(
            remetente="75.551.583/0001-48",
            data_fim=date(2015, 1, 28),
            data_inicio=date(2015, 1, 28),
            pagina=1,
            recebidas_por="04.151.050/0001-20",
        )

        assinador.assinar_xml(pedido.gerar_xml(assinador=assinador))

        cliente = ClienteMock()
        cliente.consultar_notas_periodo(pedido)

    def test_emitidas(self):
        assinador = AssinadorMock()
        self.maxDiff = None

        pedido = ConsultaNFePeriodo(
            remetente="75.551.583/0001-48",
            data_fim=date(2015, 1, 28),
            data_inicio=date(2015, 1, 28),
            pagina=1,
            inscricao_municipal="12345678",
        )

        assinador.assinar_xml(pedido.gerar_xml(assinador=assinador))

        cliente = ClienteMock()
        cliente.consultar_notas_periodo(pedido)



from abc import ABC, abstractmethod
from abstra_notas.assinatura import Assinador
from lxml.etree import ElementBase
from .templates import load_template
from jinja2 import Template


class Pedido(ABC):
    @abstractmethod
    def gerar_xml(self, assinador: Assinador) -> ElementBase:
        raise NotImplementedError

    @property
    def template(self) -> Template:
        return load_template(self.__class__.__name__)

    @property
    def metodo(self) -> str:
        return self.__class__.__name__.replace("Pedido", "")

    @property
    def classe_retorno(self):
        return f"Retorno{self.__class__.__name__}"



from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization.pkcs12 import (
    load_key_and_certificates,
)
from cryptography.hazmat.backends import default_backend
from pathlib import Path
from tempfile import NamedTemporaryFile
from lxml.etree import tostring, fromstring, ElementBase
import xmlsec


class Assinador:
    pfx_path: Path
    pfx_password: str
    cert_pem_bytes: bytes
    private_key_pem_bytes: bytes

    def __init__(
        self,
        pfx_path: Path,
        pfx_password: str,
    ):
        self.pfx_path = pfx_path
        self.pfx_password = pfx_password

        with open(self.pfx_path, "rb") as f:
            pfx_data = f.read()

        private_key, certificate, _ = load_key_and_certificates(
            pfx_data, self.pfx_password.encode(), backend=default_backend()
        )

        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        cert_pem = certificate.public_bytes(serialization.Encoding.PEM)

        self.cert_pem_bytes = cert_pem
        self.private_key_pem_bytes = private_key_pem

    @property
    def cert_pem_file(self):
        file = NamedTemporaryFile()
        file.write(self.cert_pem_bytes)
        file.seek(0)
        return file

    @property
    def private_key_pem_file(self):
        file = NamedTemporaryFile()
        file.write(self.private_key_pem_bytes)
        file.seek(0)
        return file

    def assinar_xml(self, element: ElementBase) -> ElementBase:
        element = fromstring(tostring(element, encoding=str))
        key = xmlsec.Key.from_memory(
            self.private_key_pem_bytes,
            format=xmlsec.constants.KeyDataFormatPem,
            password=self.pfx_password,
        )
        signature_node: ElementBase = xmlsec.template.create(
            element,
            c14n_method=xmlsec.constants.TransformInclC14N,
            sign_method=xmlsec.constants.TransformRsaSha1,
        )
        element.append(signature_node)
        ref = xmlsec.template.add_reference(
            signature_node, xmlsec.constants.TransformSha1, uri=""
        )
        xmlsec.template.add_transform(ref, xmlsec.constants.TransformEnveloped)
        xmlsec.template.add_transform(ref, xmlsec.constants.TransformInclC14N)
        key_info = xmlsec.template.ensure_key_info(signature_node)
        xmlsec.template.add_x509_data(key_info)
        ctx = xmlsec.SignatureContext()
        ctx.key = key
        ctx.key.load_cert_from_memory(
            self.cert_pem_bytes, xmlsec.constants.KeyDataFormatPem
        )
        ctx.sign(signature_node)
        return element

    def assinar_bytes_rsa_sh1(self, data: bytes) -> bytes:
        private_key = serialization.load_pem_private_key(
            self.private_key_pem_bytes,
            password=None,
            backend=default_backend(),
        )

        signature = private_key.sign(
            data,
            padding.PKCS1v15(),
            hashes.SHA1(),
        )
        return signature


class AssinadorMock:
    def assinar_xml(self, element: ElementBase) -> ElementBase:
        signature = """
        <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
            <SignedInfo>
                <CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315" />
                <SignatureMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315" />
                <Reference>
                    <Transforms>
                        <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature" />
                    </Transforms>
                    <DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature" />
                    <DigestValue />
                </Reference>
            </SignedInfo>
            <SignatureValue />
            <KeyInfo>
                <X509Data>
                    <X509Certificate />
                </X509Data>
            </KeyInfo>
        </Signature>
        """
        element.append(fromstring(signature))
        return element

    def assinar_bytes_rsa_sh1(self, data: bytes) -> bytes:
        return data



from .cpf import cpf_valido, normalizar_cpf
from .cnpj import cnpj_valido, normalizar_cnpj
from typing import Literal


def cpf_ou_cnpj(valor: str, optional=False) -> Literal["CPF", "CNPJ", None]:
    if valor is None and optional:
        return None
    if cpf_valido(valor):
        return "CPF"
    elif cnpj_valido(valor):
        return "CNPJ"
    else:
        raise ValueError("Valor não é um CPF ou CNPJ válido.")


def normalizar_cpf_ou_cnpj(valor: str, optional=False) -> str:
    if valor is None and optional:
        return None
    if cpf_valido(valor):
        return normalizar_cpf(valor)
    elif cnpj_valido(valor):
        return normalizar_cnpj(valor)
    else:
        raise ValueError("Valor não é um CPF ou CNPJ válido.")



def normalizar_inscricao_municipal(inscricao_municipal, optional=False):
    if inscricao_municipal is None and optional:
        return None
    if isinstance(inscricao_municipal, int):
        inscricao_municipal = str(inscricao_municipal)
    inscricao_municipal = inscricao_municipal.zfill(8)
    assert (
        len(inscricao_municipal) == 8
    ), f"A inscrição deve ter 8 caracteres. Recebido: {inscricao_municipal}"
    return inscricao_municipal


def normalizar_codigo_verificacao(codigo, optional=False):
    if codigo is None and optional:
        return None
    codigo = "".join(filter(str.isalnum, codigo)).upper()
    assert (
        codigo is None or isinstance(codigo, str) and len(codigo) == 8
    ), f"O código de verificação deve ter 8 caracteres. Recebido: {codigo}"
    return codigo



from dataclasses import dataclass


@dataclass
class Erro(Exception):
    codigo: int
    descricao: str

    def __str__(self):
        return f"Erro {self.codigo}: {self.descricao}"

    @property
    def sucesso(self):
        return False



from dataclasses import dataclass
from typing import Literal
from abstra_notas.validacoes.cpfcnpj import normalizar_cpf_ou_cnpj, cpf_ou_cnpj


@dataclass
class Remessa:
    remetente: str
    """
    CPF ou CNPJ do remetente (prestador de serviços). Qualquer formato é aceito, ex: 00000000000, 00.000.000/0000-00.
    """

    def __post_init__(self):
        self.remetente = normalizar_cpf_ou_cnpj(self.remetente)

    @property
    def remetente_tipo(self) -> Literal["CPF", "CNPJ"]:
        return cpf_ou_cnpj(self.remetente)



from .envio_rps import (
    EnvioRPS,
    RetornoEnvioRps,
    RPS,
    EnvioLoteRPS,
    RetornoEnvioRpsLote,
)
from .consulta_cnpj import ConsultaCNPJ, RetornoConsultaCNPJ
from .cancelamento_nfe import (
    CancelamentoNFe,
    RetornoCancelamentoNFe,
)
from .consulta import ConsultaNFe, RetornoConsulta
from .erro import Erro
from .cliente import Cliente

__all__ = [
    # EnvioRPS
    "EnvioRPS",
    "RetornoEnvioRps",
    # ConsultaCNPJ
    "ConsultaCNPJ",
    "RetornoConsultaCNPJ",
    # CancelamentoNFe
    "CancelamentoNFe",
    "RetornoCancelamentoNFe",
    # Cliente
    "Cliente",
    # EnvioLoteRPS
    "EnvioLoteRPS",
    "RetornoEnvioRpsLote",
    # ConsultaNFe
    "ConsultaNFe",
    "RetornoConsulta",
    # RPS
    "RPS",
    # Erro
    "Erro",
]



from abc import abstractmethod, ABC
from abstra_notas.assinatura import Assinador
from .templates import load_template
from zeep.plugins import HistoryPlugin
from zeep import Client, Transport, Settings
from requests import Session
from lxml.etree import tostring, fromstring, ElementBase
from pathlib import Path
from typing import Generic, TypeVar
from tempfile import mktemp
# import xmlschema
from xmlschema import XMLSchema
import warnings


T = TypeVar('T', bound='Envio')

class Envio(ABC, Generic[T]):
    def gerar_xml(self) -> ElementBase:

        xml = load_template(self.__class__.__name__).render(self.__dict__)
        return fromstring(xml.encode("utf-8"))
    
    @abstractmethod
    def nome_operacao(self):
        raise NotImplementedError("Subclasses must implement this method.")
    
    @abstractmethod
    def resposta(self, xml: ElementBase):
        """
        Método para processar a resposta do serviço e retornar um objeto de resposta.
        Deve ser implementado por subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method.")
    
    @property
    def schema_path(self) -> Path:
        return Path(__file__).parent / "schemas" / f"{self.nome_operacao()}.xsd"


    def executar(self, caminho_pfx: Path, senha_pfx: str, homologacao=False) -> T:
        assinador = Assinador(caminho_pfx, senha_pfx)
        try:
            history = HistoryPlugin()
            keyfile = Path(mktemp())
            keyfile.write_bytes(assinador.private_key_pem_bytes)
            certfile = Path(mktemp())
            certfile.write_bytes(assinador.cert_pem_bytes)

            xml = self.gerar_xml()
            session = Session()
            session.cert = (certfile, keyfile)
            session.verify = False
            settings = Settings(strict=True, xml_huge_tree=True)
            transport = Transport(session=session, cache=None)

            if homologacao:
                warnings.warn(
                "ATENÇÃO: Para usar o ambiente de homologação, você deve primeiro ativar "
                "o ambiente de homologação no site da Prefeitura de Fortaleza "
                "(https://iss.fortaleza.ce.gov.br/grpfor) na seção 'Controle de Acesso'.",
                UserWarning,
                stacklevel=2
            )
                url= "http://isshomo.sefin.fortaleza.ce.gov.br/grpfor-iss/ServiceGinfesImplService?wsdl"
            else:
                url = "https://iss.fortaleza.ce.gov.br/grpfor-iss/ServiceGinfesImplService?wsdl"
            client = Client(
                url, transport=transport, settings=settings, plugins=[history]
            )

            # Assina o XML
            xml_assinado = assinador.assinar_xml(xml)

            request_tmp_path = Path(mktemp())
            request_tmp_path.write_text(tostring(xml_assinado, encoding=str), encoding="utf-8")

            
            if self.nome_operacao() == "CancelarNfse":
                response: str = getattr(client.service, self.nome_operacao())(
                    tostring(xml_assinado, encoding=str)
                )
            else:
                cabecalho = cabecalho = '<?xml version="1.0" encoding="UTF-8"?><cabecalho xmlns="http://www.ginfes.com.br/cabecalho_v03.xsd" versao="3"><versaoDados>3</versaoDados></cabecalho>'
                response: str = getattr(client.service, self.nome_operacao())(
                    cabecalho, tostring(xml_assinado, encoding=str)
                )

            response_temp_path = Path(mktemp())
            response_temp_path.write_text(response, encoding="utf-8")
            #print(f"Response saved to: {response_temp_path}")
            xml_resposta =  fromstring(response.encode("utf-8"))
            return self.resposta(xml_resposta)
        finally:
            keyfile.unlink()
            certfile.unlink()


from .pedido import Pedido
from .retorno import Retorno
from dataclasses import dataclass
from lxml.etree import Element, fromstring, ElementBase, tostring
from abstra_notas.assinatura import Assinador
from typing import Literal, List
from abstra_notas.validacoes.cpfcnpj import cpf_ou_cnpj, normalizar_cpf_ou_cnpj
from .remessa import Remessa
from .erro import Erro


@dataclass
class RetornoConsultaCNPJDetalhe:
    inscricao_municipal: str
    emite_nfe: bool

    @staticmethod
    def ler_xml(xml: ElementBase) -> "RetornoConsultaCNPJDetalhe":
        return RetornoConsultaCNPJDetalhe(
            inscricao_municipal=xml.find(".//InscricaoMunicipal").text,
            emite_nfe=xml.find(".//EmiteNFe").text == "true",
        )


@dataclass
class RetornoConsultaCNPJ(Retorno):
    detalhes: List[RetornoConsultaCNPJDetalhe]

    @staticmethod
    def ler_xml(xml: ElementBase) -> "RetornoConsultaCNPJ":
        sucesso = xml.find(".//Sucesso").text == "true"
        if sucesso:
            return RetornoConsultaCNPJ(
                detalhes=[
                    RetornoConsultaCNPJDetalhe.ler_xml(detalhe)
                    for detalhe in xml.findall(".//Detalhe")
                ]
            )
        else:
            raise ErroConsultaCNPJ(
                codigo=int(xml.find(".//Codigo").text),
                descricao=xml.find(".//Descricao").text,
            )


@dataclass
class ErroConsultaCNPJ(Erro):
    codigo: int
    descricao: str


@dataclass
class ConsultaCNPJ(Pedido, Remessa):
    contribuinte: str

    def __post_init__(self):
        self.contribuinte = normalizar_cpf_ou_cnpj(self.contribuinte)

    def gerar_xml(self, assinador: Assinador) -> Element:
        xml = self.template.render(
            remetente=self.remetente,
            contribuinte=self.contribuinte,
            contribuinte_tipo=self.contribuinte_tipo,
            remetente_tipo=self.remetente_tipo,
        )
        return fromstring(xml)

    @property
    def contribuinte_tipo(self) -> Literal["CPF", "CNPJ"]:
        return cpf_ou_cnpj(self.contribuinte)



from .pedido import Pedido
from .erro import Erro
from .retorno import Retorno
from .remessa import Remessa
from abstra_notas.assinatura import Assinador
from dataclasses import dataclass
from typing import Optional, List
from lxml.etree import ElementBase, fromstring
from datetime import date
from dateutil.parser import parse
from .validacoes import normalizar_inscricao_municipal, normalizar_codigo_verificacao
from abstra_notas.validacoes.data import normalizar_data
from abstra_notas.validacoes.cpfcnpj import normalizar_cpf_ou_cnpj, cpf_ou_cnpj


def find_text(xml: ElementBase, xpath: str) -> Optional[str]:
    element = xml.find(xpath)
    return element.text.strip() if element is not None and element.text else None


def optional_int(arg: Optional[str]) -> Optional[int]:
    return int(arg) if arg is not None else None


def optional_float(arg: Optional[str]) -> Optional[float]:
    return float(arg) if arg is not None else None


def optional_bool(arg: Optional[str]) -> Optional[bool]:
    if arg is None:
        return None
    elif arg == "true":
        return True
    elif arg == "false":
        return False


def optional_centavos(arg: Optional[str]) -> Optional[int]:
    return int(optional_float(arg) * 100) if arg is not None else None


def optional_date(arg: Optional[str]) -> Optional[date]:
    return parse(arg).date() if arg is not None else None


@dataclass
class RetornoNFe:
    assinatura: str
    chave_nfe_inscricao_prestador: str
    chave_nfe_numero_nfe: int
    chave_nfe_codigo_verificacao: str
    data_emissao_nfe: date
    cpf_cnpj_prestador: str
    razao_social_prestador: str
    tipo_logradouro_prestador: str
    logradouro_prestador: str
    numero_endereco_prestador: str
    complemento_endereco_prestador: str
    bairro_prestador: str
    cidade_prestador: str
    uf_prestador: str
    cep_prestador: str
    status_nfe: str
    tributacao_nfe: str
    opcao_simples: int
    codigo_servico: int
    aliquota_servicos: float
    iss_retido: bool
    discriminacao: str
    valor_iss_centavos: int
    valor_credito_centavos: int
    valor_servicos_centavos: int
    valor_deducoes_centavos: Optional[int] = None
    valor_pis_centavos: Optional[int] = None
    valor_cofins_centavos: Optional[int] = None
    valor_inss_centavos: Optional[int] = None
    valor_ir_centavos: Optional[int] = None
    valor_csll_centavos: Optional[int] = None
    cpf_cnpj_tomador: Optional[str] = None
    razao_social_tomador: Optional[str] = None
    tipo_logradouro_tomador: Optional[str] = None
    logradouro_tomador: Optional[str] = None
    numero_endereco_tomador: Optional[str] = None
    bairro_tomador: Optional[str] = None
    cidade_tomador: Optional[str] = None
    uf_tomador: Optional[str] = None
    cep_tomador: Optional[str] = None
    email_tomador: Optional[str] = None
    chave_rps_inscricao_prestador: Optional[str] = None
    chave_rps_serie_rps: Optional[str] = None
    chave_rps_numero_rps: Optional[int] = None
    tipo_rps: Optional[str] = None
    data_emissao_rps: Optional[date] = None
    email_prestador: Optional[str] = None
    data_cancelamento: Optional[date] = None
    numero_guia: Optional[str] = None
    data_quitacao_guia: Optional[date] = None
    cpf_cnpj_intermediario: Optional[str] = None
    inscricao_municipal_intermediario: Optional[str] = None
    iss_retido_intermediario: Optional[bool] = None
    email_intermediario: Optional[str] = None
    valor_carga_tributaria_centavos: Optional[int] = None
    fonte_carga_tributaria: Optional[str] = None
    percentual_carga_triutaria: Optional[float] = None
    codigo_cei: Optional[int] = None
    matricula_obra: Optional[int] = None
    municipio_prestacao: Optional[str] = None
    numero_encapsulamento: Optional[int] = None
    valor_total_recebido_centavos: Optional[int] = None

    def __post_init__(self):
        self.fonte_carga_tributaria = self.fonte_carga_tributaria or None

    @staticmethod
    def ler_xml(xml: ElementBase):
        return RetornoNFe(
            assinatura=find_text(xml, ".//Assinatura"),
            chave_nfe_inscricao_prestador=find_text(
                xml, ".//ChaveNFe/InscricaoPrestador"
            ),
            chave_nfe_numero_nfe=optional_int(find_text(xml, ".//ChaveNFe/NumeroNFe")),
            chave_nfe_codigo_verificacao=find_text(
                xml, ".//ChaveNFe/CodigoVerificacao"
            ),
            data_emissao_nfe=optional_date(find_text(xml, ".//DataEmissaoNFe")),
            chave_rps_inscricao_prestador=find_text(
                xml, ".//ChaveRPS/InscricaoPrestador"
            ),
            chave_rps_serie_rps=find_text(xml, ".//ChaveRPS/SerieRPS"),
            chave_rps_numero_rps=optional_int(find_text(xml, ".//ChaveRPS/NumeroRPS")),
            tipo_rps=find_text(xml, ".//TipoRPS"),
            data_emissao_rps=optional_date(find_text(xml, ".//DataEmissaoRPS")),
            cpf_cnpj_prestador=find_text(xml, ".//CPFCNPJPrestador/CNPJ"),
            razao_social_prestador=find_text(xml, ".//RazaoSocialPrestador"),
            tipo_logradouro_prestador=find_text(
                xml, ".//EnderecoPrestador/TipoLogradouro"
            ),
            logradouro_prestador=find_text(xml, ".//EnderecoPrestador/Logradouro"),
            numero_endereco_prestador=find_text(
                xml, ".//EnderecoPrestador/NumeroEndereco"
            ),
            complemento_endereco_prestador=find_text(
                xml, ".//EnderecoPrestador/ComplementoEndereco"
            ),
            bairro_prestador=find_text(xml, ".//EnderecoPrestador/Bairro"),
            cidade_prestador=find_text(xml, ".//EnderecoPrestador/Cidade"),
            uf_prestador=find_text(xml, ".//EnderecoPrestador/UF"),
            cep_prestador=find_text(xml, ".//EnderecoPrestador/CEP"),
            status_nfe=find_text(xml, ".//StatusNFe"),
            tributacao_nfe=find_text(xml, ".//TributacaoNFe"),
            opcao_simples=optional_int(find_text(xml, ".//OpcaoSimples")),
            valor_servicos_centavos=optional_centavos(
                find_text(xml, ".//ValorServicos")
            ),
            valor_deducoes_centavos=optional_centavos(
                find_text(xml, ".//ValorDeducoes")
            ),
            valor_pis_centavos=optional_centavos(find_text(xml, ".//ValorPIS")),
            valor_cofins_centavos=optional_centavos(find_text(xml, ".//ValorCOFINS")),
            valor_inss_centavos=optional_centavos(find_text(xml, ".//ValorINSS")),
            valor_ir_centavos=optional_centavos(find_text(xml, ".//ValorIR")),
            valor_csll_centavos=optional_centavos(find_text(xml, ".//ValorCSLL")),
            valor_iss_centavos=optional_centavos(find_text(xml, ".//ValorISS")),
            valor_credito_centavos=optional_centavos(find_text(xml, ".//ValorCredito")),
            codigo_servico=optional_int(find_text(xml, ".//CodigoServico")),
            aliquota_servicos=optional_float(find_text(xml, ".//AliquotaServicos")),
            iss_retido=optional_bool(find_text(xml, ".//ISSRetido")),
            cpf_cnpj_tomador=find_text(xml, ".//CPFCNPJTomador/CPF")
            or find_text(xml, ".//CPFCNPJTomador/CNPJ"),
            razao_social_tomador=find_text(xml, ".//RazaoSocialTomador"),
            tipo_logradouro_tomador=find_text(xml, ".//EnderecoTomador/TipoLogradouro"),
            logradouro_tomador=find_text(xml, ".//EnderecoTomador/Logradouro"),
            numero_endereco_tomador=find_text(xml, ".//EnderecoTomador/NumeroEndereco"),
            bairro_tomador=find_text(xml, ".//EnderecoTomador/Bairro"),
            cidade_tomador=find_text(xml, ".//EnderecoTomador/Cidade"),
            uf_tomador=find_text(xml, ".//EnderecoTomador/UF"),
            cep_tomador=find_text(xml, ".//EnderecoTomador/CEP"),
            email_tomador=find_text(xml, ".//EmailTomador"),
            discriminacao=find_text(xml, ".//Discriminacao"),
            fonte_carga_tributaria=find_text(xml, ".//FonteCargaTributaria"),
        )


@dataclass
class RetornoConsulta(Retorno):
    lista_nfe: List[RetornoNFe]

    @staticmethod
    def ler_xml(xml: ElementBase):
        sucesso = xml.find(".//Sucesso").text == "true"
        if sucesso:
            lista_nfe = []
            for nfe_xml in xml.findall(".//NFe"):
                lista_nfe.append(RetornoNFe.ler_xml(nfe_xml))

            return RetornoConsulta(lista_nfe=lista_nfe)
        else:
            raise Erro(
                codigo=int(xml.find(".//Codigo").text),
                descricao=xml.find(".//Descricao").text,
            )


@dataclass
class ConsultaNFe(Pedido, Remessa):
    chave_nfe_inscricao_prestador: str
    chave_nfe_numero_nfe: int
    chave_rps_inscricao_prestador: str
    chave_rps_serie_rps: str
    chave_rps_numero_rps: int
    chave_nfe_codigo_verificacao: Optional[str] = None
    """
    Código de verificação da NFe gerada pelo sistema de notas fiscais eletrônicas.
    """

    def __post_init__(self):
        self.chave_nfe_inscricao_prestador = normalizar_inscricao_municipal(
            self.chave_nfe_inscricao_prestador
        )
        self.chave_rps_inscricao_prestador = normalizar_inscricao_municipal(
            self.chave_rps_inscricao_prestador
        )
        self.chave_nfe_codigo_verificacao = normalizar_codigo_verificacao(
            self.chave_nfe_codigo_verificacao, optional=True
        )

    def gerar_xml(self, assinador: Assinador):
        xml = self.template.render(
            remetente=self.remetente,
            remetente_tipo=self.remetente_tipo,
            chave_nfe_inscricao_prestador=self.chave_nfe_inscricao_prestador,
            chave_nfe_numero_nfe=self.chave_nfe_numero_nfe,
            chave_rps_inscricao_prestador=self.chave_rps_inscricao_prestador,
            chave_rps_serie_rps=self.chave_rps_serie_rps,
            chave_rps_numero_rps=self.chave_rps_numero_rps,
            chave_nfe_codigo_verificacao=self.chave_nfe_codigo_verificacao,
        )
        return fromstring(xml.encode("utf-8"))

    @property
    def classe_retorno(self):
        return RetornoConsulta.__name__


@dataclass
class ConsultaNFePeriodo(Pedido, Remessa):
    """
    Consultar notas fiscais eletrônicas emitidas e/ou recebidas em um período.

    Ao menos um dos parâmetros `recebidas_por` ou `inscricao_municipal` deve ser informado.
    """

    data_inicio: date
    data_fim: date

    inscricao_municipal: Optional[str] = None
    """
    Parâmetro opcional. (Inscrição Municipal, formatação opcional, ex: 12345678)

    Caso seja informado, a consulta será feita apenas para as notas fiscais emitidas pelo remetente informado.
    """

    recebidas_por: Optional[str] = None
    """
    Parâmetro opcional. (CPF ou CNPJ, formatação opcional, ex: 12345678901 ou 12345678000123)
    
    Caso seja informado, a consulta será feita apenas para as notas fiscais recebidas pelo tomador informado.

    Caso não seja informado, a consulta será feita para todas as notas fiscais emitidas pelo remetente.
    """

    pagina: int = 1

    def __post_init__(self):
        self.inscricao_municipal = normalizar_inscricao_municipal(
            self.inscricao_municipal, optional=True
        )
        self.data_inicio = normalizar_data(self.data_inicio)
        self.data_fim = normalizar_data(self.data_fim)
        self.recebidas_por = normalizar_cpf_ou_cnpj(self.recebidas_por, optional=True)
        self.remetente = normalizar_cpf_ou_cnpj(self.remetente)
        self.pagina = int(self.pagina)

        if self.inscricao_municipal is None and self.recebidas_por is None:
            raise ValueError(
                "Ao menos um dos parâmetros `recebidas_por` ou `inscricao_municipal` deve ser informado."
            )
        elif self.inscricao_municipal is not None and self.recebidas_por is not None:
            raise ValueError(
                "Os parâmetros `recebidas_por` e `inscricao_municipal` são mutuamente exclusivos."
            )

    @property
    def recebidas_por_tipo(self):
        return cpf_ou_cnpj(self.recebidas_por, optional=True)

    def gerar_xml(self, assinador: Assinador):
        xml = self.template.render(
            remetente=self.remetente,
            remetente_tipo=self.remetente_tipo,
            inscricao_municipal=self.inscricao_municipal,
            data_inicio=self.data_inicio,
            data_fim=self.data_fim,
            pagina=self.pagina,
            recebidas_por=self.recebidas_por,
            recebidas_por_tipo=self.recebidas_por_tipo,
        )
        return fromstring(xml.encode("utf-8"))

    @property
    def classe_retorno(self):
        return RetornoConsulta.__name__



from pathlib import Path
from jinja2 import Template

current_path = Path(__file__).parent


def load_template(class_name):
    with open(current_path / f"{class_name}.xml", "r", encoding="utf-8") as file:
        return Template(file.read())



from typing import Union


def normalizar_cep(cep: Union[str, int]) -> str:
    cep = str(cep)
    cep = cep.replace("-", "").replace(".", "")
    assert cep.isdigit(), "CEP deve conter apenas números"
    cep = cep.zfill(8)
    assert len(cep) == 8, "CEP deve conter 8 dígitos"
    return cep



#!/usr/bin/env python3
"""
Testes unitários para o parsing da resposta do EnviarLoteRpsEnvio
Usa respostas mock para testar sem fazer requests reais.
"""

import unittest
from datetime import datetime
from lxml.etree import fromstring
from abstra_notas.nfse.ce.fortaleza import EnviarLoteRpsEnvio


class TestEnviarLoteRpsResponseParsing(unittest.TestCase):
    """Testes para o parsing das respostas do EnviarLoteRpsEnvio"""

    def setUp(self):
        """Configuração comum para todos os testes"""
        self.envio = EnviarLoteRpsEnvio(
            lote_id="LOTE001",
            numero_lote="123", 
            prestador_cnpj="11.222.333/0001-81",
            prestador_inscricao_municipal="123456",
            lista_rps=[]
        )

    def test_response_parsing_success(self):
        """Testa o parsing de uma resposta de sucesso"""
        # XML de resposta de sucesso mock
        response_xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <EnviarLoteRpsResposta xmlns="http://www.ginfes.com.br/servico_enviar_lote_rps_resposta_v03.xsd" xmlns:ns2="http://www.ginfes.com.br/tipos_v03.xsd">
            <NumeroLote>123</NumeroLote>
            <DataRecebimento>2024-01-15T14:30:25.123-03:00</DataRecebimento>
            <Protocolo>987654321</Protocolo>
        </EnviarLoteRpsResposta>"""
        
        # Parse do XML
        xml_resposta = fromstring(response_xml.encode("utf-8"))
        
        # Executar parsing
        resultado = self.envio.resposta(xml_resposta)
        
        # Validações
        self.assertIsNotNone(resultado)
        self.assertEqual(resultado.numero_lote, "123")
        self.assertEqual(resultado.protocolo, "987654321")
        self.assertIsNone(resultado.lista_mensagem_retorno)
        self.assertIsInstance(resultado.data_recebimento, datetime)
        
        # Verificar que a data foi parseada corretamente
        expected_date = datetime.fromisoformat("2024-01-15T14:30:25.123-03:00")
        self.assertEqual(resultado.data_recebimento, expected_date)

    def test_response_parsing_error(self):
        """Testa o parsing de uma resposta com erro"""
        # XML de resposta com erro mock
        response_xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <EnviarLoteRpsResposta xmlns="http://www.ginfes.com.br/servico_enviar_lote_rps_resposta_v03.xsd" xmlns:ns2="http://www.ginfes.com.br/tipos_v03.xsd">
            <ns2:ListaMensagemRetorno>
                <ns2:MensagemRetorno>
                    <ns2:Codigo>E999</ns2:Codigo>
                    <ns2:Mensagem>Erro de teste para validação.</ns2:Mensagem>
                    <ns2:Correcao>Corrigir dados de teste.</ns2:Correcao>
                </ns2:MensagemRetorno>
            </ns2:ListaMensagemRetorno>
        </EnviarLoteRpsResposta>"""
        
        # Parse do XML
        xml_resposta = fromstring(response_xml.encode("utf-8"))
        
        # Executar parsing
        resultado = self.envio.resposta(xml_resposta)
        
        # Validações
        self.assertIsNotNone(resultado)
        self.assertIsNone(resultado.numero_lote)
        self.assertIsNone(resultado.data_recebimento)
        self.assertIsNone(resultado.protocolo)
        self.assertIsNotNone(resultado.lista_mensagem_retorno)
        self.assertEqual(len(resultado.lista_mensagem_retorno), 1)
        
        # Verificar mensagem de erro
        erro = resultado.lista_mensagem_retorno[0]
        self.assertEqual(erro.codigo, "E999")
        self.assertEqual(erro.mensagem, "Erro de teste para validação.")
        self.assertEqual(erro.correcao, "Corrigir dados de teste.")

    def test_response_parsing_multiple_errors(self):
        """Testa o parsing de uma resposta com múltiplos erros"""
        # XML de resposta com múltiplos erros mock
        response_xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <EnviarLoteRpsResposta xmlns="http://www.ginfes.com.br/servico_enviar_lote_rps_resposta_v03.xsd" xmlns:ns2="http://www.ginfes.com.br/tipos_v03.xsd">
            <ns2:ListaMensagemRetorno>
                <ns2:MensagemRetorno>
                    <ns2:Codigo>E001</ns2:Codigo>
                    <ns2:Mensagem>Primeiro erro de teste.</ns2:Mensagem>
                    <ns2:Correcao>Corrigir primeiro problema.</ns2:Correcao>
                </ns2:MensagemRetorno>
                <ns2:MensagemRetorno>
                    <ns2:Codigo>E002</ns2:Codigo>
                    <ns2:Mensagem>Segundo erro de teste.</ns2:Mensagem>
                    <ns2:Correcao>Corrigir segundo problema.</ns2:Correcao>
                </ns2:MensagemRetorno>
            </ns2:ListaMensagemRetorno>
        </EnviarLoteRpsResposta>"""
        
        # Parse do XML
        xml_resposta = fromstring(response_xml.encode("utf-8"))
        
        # Executar parsing
        resultado = self.envio.resposta(xml_resposta)
        
        # Validações
        self.assertIsNotNone(resultado)
        self.assertIsNone(resultado.numero_lote)
        self.assertIsNone(resultado.data_recebimento)
        self.assertIsNone(resultado.protocolo)
        self.assertIsNotNone(resultado.lista_mensagem_retorno)
        self.assertEqual(len(resultado.lista_mensagem_retorno), 2)
        
        # Verificar primeira mensagem de erro
        erro1 = resultado.lista_mensagem_retorno[0]
        self.assertEqual(erro1.codigo, "E001")
        self.assertEqual(erro1.mensagem, "Primeiro erro de teste.")
        
        # Verificar segunda mensagem de erro
        erro2 = resultado.lista_mensagem_retorno[1]
        self.assertEqual(erro2.codigo, "E002")
        self.assertEqual(erro2.mensagem, "Segundo erro de teste.")

    def test_response_type_validation(self):
        """Testa se o tipo de resposta retornado é correto"""
        response_xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <EnviarLoteRpsResposta xmlns="http://www.ginfes.com.br/servico_enviar_lote_rps_resposta_v03.xsd" xmlns:ns2="http://www.ginfes.com.br/tipos_v03.xsd">
            <NumeroLote>456</NumeroLote>
            <DataRecebimento>2024-02-20T09:15:30.456-03:00</DataRecebimento>
            <Protocolo>123456789</Protocolo>
        </EnviarLoteRpsResposta>"""
        
        # Parse do XML
        xml_resposta = fromstring(response_xml.encode("utf-8"))
        
        # Executar parsing
        resultado = self.envio.resposta(xml_resposta)
        
        # Validar tipo da resposta
        from abstra_notas.nfse.ce.fortaleza import EnviarLoteRpsResposta
        self.assertIsInstance(resultado, EnviarLoteRpsResposta)

    def test_empty_response_handling(self):
        """Testa o comportamento com uma resposta mínima válida"""
        response_xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <EnviarLoteRpsResposta xmlns="http://www.ginfes.com.br/servico_enviar_lote_rps_resposta_v03.xsd" xmlns:ns2="http://www.ginfes.com.br/tipos_v03.xsd">
        </EnviarLoteRpsResposta>"""
        
        # Parse do XML
        xml_resposta = fromstring(response_xml.encode("utf-8"))
        
        # Executar parsing
        resultado = self.envio.resposta(xml_resposta)
        
        # Validações - campos devem ser None quando não presentes
        self.assertIsNotNone(resultado)
        self.assertIsNone(resultado.numero_lote)
        self.assertIsNone(resultado.data_recebimento)
        self.assertIsNone(resultado.protocolo)
        self.assertIsNone(resultado.lista_mensagem_retorno)


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from typing import Literal, List, Optional
from lxml.etree import Element, fromstring, ElementBase
import base64
from dateutil.parser import parse
from abstra_notas.validacoes.email import validar_email
from abstra_notas.validacoes.cidades import validar_codigo_cidade, normalizar_uf, UF
from abstra_notas.validacoes.cpfcnpj import normalizar_cpf_ou_cnpj, cpf_ou_cnpj
from abstra_notas.validacoes.cep import normalizar_cep
from abstra_notas.validacoes.tipo_logradouro import TipoLogradouro
from datetime import date
from .remessa import Remessa
from .pedido import Pedido
from .retorno import Retorno
from .templates import load_template
from abstra_notas.assinatura import Assinador
from .erro import Erro
import re
import re


@dataclass
class ChaveNFeRPS(Retorno):
    chave_nfe_inscricao_prestador: str
    chave_nfe_numero_nfe: int
    chave_nfe_codigo_verificacao: str
    chave_rps_inscricao_prestador: str
    chave_rps_serie_rps: str
    chave_rps_numero_rps: int

    def __post_init__(self):
        if isinstance(self.chave_nfe_inscricao_prestador, int):
            self.chave_nfe_inscricao_prestador = str(self.chave_nfe_inscricao_prestador)

        if isinstance(self.chave_rps_inscricao_prestador, int):
            self.chave_rps_inscricao_prestador = str(self.chave_rps_inscricao_prestador)

    @staticmethod
    def ler_xml(xml: ElementBase) -> "ChaveNFeRPS":
        return ChaveNFeRPS(
            chave_nfe_inscricao_prestador=xml.find(".//ChaveNFe")
            .find(".//InscricaoPrestador")
            .text,
            chave_nfe_codigo_verificacao=xml.find(".//ChaveNFe")
            .find(".//CodigoVerificacao")
            .text,
            chave_nfe_numero_nfe=int(xml.find(".//ChaveNFe").find(".//NumeroNFe").text),
            chave_rps_inscricao_prestador=xml.find(".//ChaveRPS")
            .find(".//InscricaoPrestador")
            .text,
            chave_rps_numero_rps=int(xml.find(".//ChaveRPS").find(".//NumeroRPS").text),
            chave_rps_serie_rps=xml.find(".//ChaveRPS").find(".//SerieRPS").text,
        )


@dataclass
class RetornoEnvioRps(ChaveNFeRPS):
    @staticmethod
    def ler_xml(xml: ElementBase) -> "RetornoEnvioRps":
        sucesso = xml.find(".//Sucesso").text
        if sucesso == "false":
            raise ErroEnvioRps(
                codigo=xml.find(".//Codigo").text,
                descricao=xml.find(".//Descricao").text,
            )
        elif sucesso == "true":
            chave = ChaveNFeRPS.ler_xml(xml)
            return RetornoEnvioRps(**chave.__dict__)


@dataclass
class ErroEnvioRps(Erro):
    codigo: int
    descricao: str


@dataclass
class RPS:
    inscricao_prestador: int
    """
    Inscrição Municipal do Prestador.
    """

    numero_rps: int
    """
    Número que identifica o RPS. Deve ser único para cada série de RPS.
    Recomenda-se que seja sequencial, iniciando em 1.
    Use algum banco de dados para salvar o número do RPS e garantir que seja único.
    """

    tipo_rps: Literal["RPS", "RPS-M", "RPS-C"]
    """
    RPS: Recibo Provisório de Serviços.
    
    RPS-M: Recibo Provisório de Serviços proveniente de Nota Fiscal Conjugada (Mista).
    
    RPS-C: Recibo Provisório de Serviços proveniente de Nota Fiscal Conjugada (Comum).
    """

    data_emissao: date
    discriminacao: str
    status_rps: Literal["N", "C"]
    tributacao_rps: Literal["T", "F", "A", "B", "D", "M", "N", "R", "S", "X", "V", "P"]
    """
    T: Tributado em São Paulo

    F: Tributado Fora de São Paulo
    
    A: Tributado em São Paulo, porém Isento
    
    B: Tributado Fora de São Paulo, porém Isento
    
    D: Tributado em São Paulo com isenção parcial
    
    M: Tributado em São Paulo, porém com indicação de imunidade subjetiva
    
    N: Tributado Fora de São Paulo, porém com indicação de imunidade subjetiva
    
    R: Tributado em São Paulo, porém com indicação de imunidade objetiva
    
    S: Tributado fora de São Paulo, porém com indicação de imunidade objetiva
    
    X: Tributado em São Paulo, porém Exigibilidade Suspensa
    
    V: Tributado Fora de São Paulo, porém Exigibilidade Suspensa
    
    P: Exportação de Serviços
    """

    codigo_servico: str
    """
    Informe o código do serviço do RPS. Este código deve pertencer à lista de serviços.

    Código de serviço com 4 ou 5 dígitos.
    """

    aliquota_servicos: float
    """
    Valor percentual da alíquota de serviços entre 0 e 1.
    """
    iss_retido: bool

    valor_servicos_centavos: int
    valor_deducoes_centavos: int

    serie_rps: str
    """
    Série do RPS com 5 posições (caracteres). Completar com espaços em branco à direita caso seja necessário.

    Atenção: Não utilize espaços à esquerda. O conteúdo deverá estar alinhado à esquerda. 
    """

    valor_pis_centavos: Optional[int] = None
    valor_cofins_centavos: Optional[int] = None
    valor_inss_centavos: Optional[int] = None
    valor_ir_centavos: Optional[int] = None
    valor_csll_centavos: Optional[int] = None
    valor_carga_tributaria_centavos: Optional[int] = None
    valor_total_recebido_centavos: Optional[int] = None

    tomador: Optional[str] = None
    """
    CPF ou CNPJ do tomador. Qualquer formato é aceito, ex: 00000000000, 00.000.000/0000-00.
    """
    inscricao_municipal_tomador: Optional[str] = None
    """
    Este elemento só deverá ser preenchido para tomadores no município de São Paulo (CCM).
    
    CCM: Cadastro de Contribuintes Mobiliários.
    """

    inscricao_estadual_tomador: Optional[str] = None

    razao_social_tomador: Optional[str] = None

    email_tomador: Optional[str] = None

    endereco_tipo_logradouro: Optional[TipoLogradouro] = None
    endereco_logradouro: Optional[str] = None
    endereco_numero: Optional[str] = None
    endereco_complemento: Optional[str] = None
    endereco_uf: Optional[UF] = None
    endereco_cidade: Optional[int] = None
    """
    Código da cidade. Pode ser obtido em:

    https://servicodados.ibge.gov.br/api/v1/localidades/municipios
    """
    endereco_bairro: Optional[str] = None
    endereco_cep: Optional[str] = None

    intermediario: Optional[str] = None
    """
    CPF ou CNPJ do intermediário. Qualquer formato é aceito, ex: 00000000000, 00.000.000/0000-00.
    """

    inscricao_municipal_intermediario: Optional[str] = None
    """
    Este elemento só deverá ser preenchido para intermediários no município de São Paulo (CCM).
    """

    iss_retido_intermediario: Optional[bool] = None
    email_intermediario: Optional[str] = None

    percentual_carga_tributaria: Optional[float] = None
    """
    Valor percentual da carga tributária entre 0 e 1.
    """

    fonte_carga_tributaria: Optional[str] = None
    """
    Fonte da carga tributária. Máximo de 10 caracteres.
    """

    codigo_cei: Optional[int] = None
    """
    Cadastro Específico do INSS.
    """

    matricula_obra: Optional[int] = None
    """
    Número de matrícula de obra.
    """

    municipio_prestacao: Optional[int] = None
    """
    Código do município onde o serviço foi prestado.

    Codigo do municipio pode ser obtido em:

    https://servicodados.ibge.gov.br/api/v1/localidades/municipios
    """

    numero_encapsulamento: Optional[int] = None
    """
    Código de encapsulamento de notas dedutoras.
    """

    def __post_init__(self):
        if isinstance(self.data_emissao, str):
            self.data_emissao = parse(self.data_emissao).date()

        if self.endereco_cep is not None:
            self.endereco_cep = normalizar_cep(self.endereco_cep)

        assert (
            self.endereco_bairro is None or len(self.endereco_bairro) <= 30
        ), "O bairro deve ter no máximo 30 caracteres"

        if self.endereco_uf is not None:
            if isinstance(self.endereco_uf, str):
                uf_str = self.endereco_uf
            else:
                uf_str = self.endereco_uf.value
            self.endereco_uf = UF(normalizar_uf(uf_str))

        if self.tomador is not None:
            self.tomador = normalizar_cpf_ou_cnpj(self.tomador)

        if self.endereco_cidade is not None:
            assert validar_codigo_cidade(
                self.endereco_cidade
            ), f"Código de cidade inválido: {self.endereco_cidade}"

        if isinstance(self.endereco_tipo_logradouro, str):
            if self.endereco_tipo_logradouro.upper() in TipoLogradouro.__members__:
                self.endereco_tipo_logradouro = TipoLogradouro[
                    self.endereco_tipo_logradouro.upper()
                ]
            else:
                self.endereco_tipo_logradouro = TipoLogradouro(
                    self.endereco_tipo_logradouro.upper()
                )

        assert (
            self.aliquota_servicos >= 0 and self.aliquota_servicos <= 1
        ), "A alíquota de serviços deve ser um valor entre 0 e 1"

        if isinstance(self.codigo_servico, int):
            self.codigo_servico = str(self.codigo_servico).zfill(4)
        elif isinstance(self.codigo_servico, str):
            assert len(self.codigo_servico) in [
                4,
                5,
            ], "Código de serviço deve ter 4 ou 5 dígitos"

        # assert (
        #     self.codigo_servico in codigos_de_servico_validos
        # ), f"Código de serviço inválido, os códigos válidos são: {codigos_de_servico_validos}"
        assert self.email_tomador is None or validar_email(
            self.email_tomador
        ), f"Email do tomador com formato inválido: {self.email_tomador}"
        assert (
            isinstance(self.valor_servicos_centavos, int)
            and self.valor_servicos_centavos >= 0
        ), "O valor de serviços deve ser um valor decimal maior ou igual a zero"
        assert self.valor_deducoes_centavos is None or (
            isinstance(self.valor_deducoes_centavos, int)
            and self.valor_deducoes_centavos >= 0
        ), "O valor de deduções deve ser um valor decimal maior ou igual a zero"
        assert self.valor_pis_centavos is None or (
            isinstance(self.valor_pis_centavos, int) and self.valor_pis_centavos >= 0
        ), "O valor de PIS deve ser um valor decimal maior ou igual a zero"
        assert self.valor_cofins_centavos is None or (
            isinstance(self.valor_cofins_centavos, int)
            and self.valor_cofins_centavos >= 0
        ), "O valor de COFINS deve ser um valor decimal maior ou igual a zero"
        assert self.valor_inss_centavos is None or (
            isinstance(self.valor_inss_centavos, int) and self.valor_inss_centavos >= 0
        ), "O valor de INSS deve ser um valor decimal maior ou igual a zero"
        assert self.valor_ir_centavos is None or (
            isinstance(self.valor_ir_centavos, int) and self.valor_ir_centavos >= 0
        ), "O valor de IR deve ser um valor decimal maior ou igual a zero"
        assert self.valor_csll_centavos is None or (
            isinstance(self.valor_csll_centavos, int) and self.valor_csll_centavos >= 0
        ), "O valor de CSLL deve ser um valor decimal maior ou igual a zero"
        assert self.valor_carga_tributaria_centavos is None or (
            isinstance(self.valor_carga_tributaria_centavos, int)
            and self.valor_carga_tributaria_centavos >= 0
        ), "O valor da carga tributária deve ser um valor decimal maior ou igual a zero"
        assert self.valor_total_recebido_centavos is None or (
            isinstance(self.valor_total_recebido_centavos, int)
            and self.valor_total_recebido_centavos >= 0
        ), "O valor total recebido deve ser um valor decimal maior ou igual a zero"
        assert (
            self.valor_carga_tributaria_centavos is None
            or self.valor_carga_tributaria_centavos >= 0
        ), "O valor da carga tributária deve ser maior ou igual a zero"
        assert (
            self.valor_servicos_centavos
            - (self.valor_deducoes_centavos or 0)
            - (self.valor_pis_centavos or 0)
            - (self.valor_cofins_centavos or 0)
            - (self.valor_inss_centavos or 0)
            - (self.valor_ir_centavos or 0)
            - (self.valor_csll_centavos or 0)
            >= 0
        ), "A soma dos valores não pode ser negativa"

        if self.serie_rps is not None:
            self.serie_rps = self.serie_rps.strip()
            assert (
                len(self.serie_rps) <= 5
            ), "A série do RPS deve ter no máximo 5 caracteres"

        if self.inscricao_municipal_tomador is not None:
            assert (
                self.inscricao_municipal_tomador.isdigit()
                and len(self.inscricao_municipal_tomador) == 8
            ), "Inscrição municipal do tomador deve ter 8 dígitos"
            assert (
                self.tomador is not None
            ), "Ao preencher a inscrição municipal do tomador, o CPF ou CNPJ do tomador deve ser preenchido"

        if self.inscricao_municipal_intermediario is not None:
            assert (
                self.inscricao_municipal_intermediario.isdigit()
                and len(self.inscricao_municipal_intermediario) == 8
            ), "Inscrição municipal do intermediário deve ter 8 dígitos"
            assert (
                self.intermediario is not None
            ), "Ao preencher a inscrição municipal do intermediário, o CPF ou CNPJ do intermediário deve ser preenchido"

        if self.intermediario is not None:
            assert (
                self.iss_retido_intermediario is not None
            ), "Ao preencher o intermediário, o ISS retido do intermediário deve ser preenchido"

        assert self.email_intermediario is None or validar_email(
            self.email_intermediario
        ), f"Email do intermediário com formato inválido: {self.email_intermediario}"

        assert (
            self.percentual_carga_tributaria is None
            or isinstance(self.percentual_carga_tributaria, float)
            and self.percentual_carga_tributaria >= 0
            and self.percentual_carga_tributaria <= 1
        ), "O percentual da carga tributária deve ser um valor entre 0 e 1"

        assert (
            self.fonte_carga_tributaria is None
            or isinstance(self.fonte_carga_tributaria, str)
            and len(self.fonte_carga_tributaria) <= 10
        ), "A fonte da carga tributária deve ter no máximo 10 caracteres"

    def gerar_string_xml(self, assinador: Assinador) -> str:
        template = load_template("RPS")
        return template.render(
            inscricao_prestador=str(self.inscricao_prestador).zfill(8),
            serie_rps=self.serie_rps,
            numero_rps=self.numero_rps,
            tipo_rps=self.tipo_rps,
            data_emissao=self.data_emissao,
            status_rps=self.status_rps,
            inscricao_estadual_tomador=self.inscricao_estadual_tomador,
            tributacao_rps=self.tributacao_rps,
            valor_servicos=f"{self.valor_servicos_centavos / 100:.2f}",
            valor_deducoes=f"{self.valor_deducoes_centavos / 100:.2f}"
            if self.valor_deducoes_centavos is not None
            else None,
            valor_pis=f"{self.valor_pis_centavos / 100:.2f}"
            if self.valor_pis_centavos is not None
            else None,
            valor_cofins=f"{self.valor_cofins_centavos / 100:.2f}"
            if self.valor_cofins_centavos is not None
            else None,
            valor_inss=f"{self.valor_inss_centavos / 100:.2f}"
            if self.valor_inss_centavos is not None
            else None,
            valor_ir=f"{self.valor_ir_centavos / 100:.2f}"
            if self.valor_ir_centavos is not None
            else None,
            valor_csll=f"{self.valor_csll_centavos / 100:.2f}"
            if self.valor_csll_centavos is not None
            else None,
            codigo_servico=self.codigo_servico,
            aliquota_servicos=self.aliquota_servicos,
            iss_retido=str(self.iss_retido).lower(),
            tomador=self.tomador,
            tomador_tipo=self.tomador_tipo,
            razao_social_tomador=self.razao_social_tomador,
            endereco_tipo_logradouro=self.endereco_tipo_logradouro.value.capitalize()
            if self.endereco_tipo_logradouro is not None
            else None,
            endereco_logradouro=self.endereco_logradouro,
            endereco_numero=self.endereco_numero,
            endereco_complemento=self.endereco_complemento,
            endereco_bairro=self.endereco_bairro,
            endereco_cidade=self.endereco_cidade,
            endereco_uf=self.endereco_uf.value
            if self.endereco_uf is not None
            else None,
            endereco_cep=self.endereco_cep,
            email_tomador=self.email_tomador,
            discriminacao=self.discriminacao,
            assinatura=self.assinatura(assinador),
            intermediario=self.intermediario,
            intermediario_tipo=self.intermediario_tipo,
            inscricao_municipal_intermediario=self.inscricao_municipal_intermediario,
            iss_retido_intermediario=self.iss_retido_intermediario,
            email_intermediario=self.email_intermediario,
            percentual_carga_tributaria=self.percentual_carga_tributaria,
            fonte_carga_tributaria=self.fonte_carga_tributaria,
            codigo_cei=self.codigo_cei,
            matricula_obra=self.matricula_obra,
            municipio_prestacao=self.municipio_prestacao,
            numero_encapsulamento=self.numero_encapsulamento,
            inscricao_municipal_tomador=self.inscricao_municipal_tomador,
            valor_carga_tributaria=f"{self.valor_carga_tributaria_centavos / 100:.2f}"
            if self.valor_carga_tributaria_centavos is not None
            else None,
            valor_total_recebido=f"{self.valor_total_recebido_centavos / 100:.2f}"
            if self.valor_total_recebido_centavos is not None
            else None,
        )

    @staticmethod
    def ler_txt(conteudo: str) -> List["RPS"]:
        lista_rps = []
        lines = conteudo.splitlines()
        inscricao_prestador = None

        for line in lines:
            if line.startswith("1"):
                inscricao_prestador = int(line[4:12])
            elif line.startswith("6"):
                if inscricao_prestador is None:
                    continue

                tipo_rps = line[1:6].strip()
                serie_rps = line[6:11].strip()
                numero_rps = int(line[11:23])
                data_emissao = parse(line[23:31]).date()
                status_rps = line[31:32]

                valor_servicos_centavos = int(line[32:47])
                valor_deducoes_centavos = int(line[47:62])

                codigo_servico = line[62:67].strip()

                aliquota_servicos = int(line[67:71]) / 10000

                iss_retido = line[71:72] == "1"
                tipo_tomador = line[72:73]

                tomador_cpf_cnpj = line[73:87].strip()

                # If Type 1 (CPF), ensure it is treated as 11 digits by stripping leading zeros/padding
                if tipo_tomador == "1":
                    # Taking the last 11 digits is safer if it was left-padded with zeros
                    if len(tomador_cpf_cnpj) > 11:
                        tomador_cpf_cnpj = tomador_cpf_cnpj[-11:]

                try:
                    tomador_cpf_cnpj = normalizar_cpf_ou_cnpj(tomador_cpf_cnpj)
                except ValueError:
                    tomador_cpf_cnpj = None

                im_tomador = line[87:95].strip()
                ie_tomador = line[95:107].strip()

                razao_social_tomador = line[107:182].strip()

                endereco_tipo_logradouro_raw = line[182:185].strip().rstrip(".")
                try:
                    if (
                        endereco_tipo_logradouro_raw.upper()
                        in TipoLogradouro.__members__
                    ):
                        endereco_tipo_logradouro = TipoLogradouro[
                            endereco_tipo_logradouro_raw.upper()
                        ]
                    else:
                        endereco_tipo_logradouro = TipoLogradouro(
                            endereco_tipo_logradouro_raw.upper()
                        )
                except ValueError:
                    endereco_tipo_logradouro = None
                endereco_logradouro = line[185:235].strip()
                endereco_numero = line[235:245].strip()
                endereco_complemento = line[245:275].strip()
                endereco_bairro = line[275:305].strip()

                # Cidade (Nome) is at 306-355, but we don't map it to IBGE int code here
                # because we don't have a lookup table loaded.
                # We will rely on 'Município Prestação' (570-576) if available or logic elsewhere.

                endereco_uf = line[355:357].strip()
                endereco_cep = line[357:365].strip()

                email_tomador_raw = line[365:440].strip()
                email_tomador = (
                    email_tomador_raw if "@" in email_tomador_raw else None
                )

                # Layout V.002 Fields (Positions 441+)
                valor_pis_centavos = int(line[440:455])
                valor_cofins_centavos = int(line[455:470])
                valor_inss_centavos = int(line[470:485])
                valor_ir_centavos = int(line[485:500])
                valor_csll_centavos = int(line[500:515])

                valor_carga_tributaria_centavos = int(line[515:530])
                percentual_carga_tributaria = int(line[530:535]) / 10000
                fonte_carga_tributaria = line[535:545].strip()

                codigo_cei = line[545:557].strip()
                matricula_obra = line[557:569].strip()
                municipio_prestacao = line[569:576].strip()
                numero_encapsulamento = line[576:586].strip()
                
                # Reserved 587-596

                valor_total_recebido_centavos = line[596:611].strip()

                # Reserved 612-786

                discriminacao = line[786:].strip()

                rps = RPS(
                    inscricao_prestador=inscricao_prestador,
                    numero_rps=numero_rps,
                    tipo_rps=tipo_rps,
                    data_emissao=data_emissao,
                    status_rps=status_rps,
                    tributacao_rps="T",  # Defaulting or needs extraction?
                    # WARNING: 'tributacao_rps' was missing in V.002 extraction in previous code too?
                    # In V.001/V.002 header logic:
                    # Pos 32: Situação do RPS (T, F, A, B...) -> This maps to 'tributacao_rps' in RPS dataclass?
                    # In RPS dataclass: status_rps is Literal["N", "C"] (Normal, Cancelado).
                    # But in Manual: Pos 32 is "Situação do RPS" with values T, F, A, B...
                    # Wait. Manual says:
                    # 6) Situação do RPS (32-32): T, F, A, B... (Taxation Type).
                    # BUT RPS Dataclass has `status_rps` AND `tributacao_rps`.
                    # Let's check RPS Dataclass:
                    # `status_rps`: Literal["N", "C"]
                    # `tributacao_rps`: Literal["T", "F", ...]
                    #
                    # In Manual V.001/V.002:
                    # Field 6 "Situação do RPS" contains T, F, A, B, C (Cancelado)...
                    # So if it is 'C', it is Cancelled. If it is 'T', it is Normal & Taxed inside SP.
                    # The `RPS` object separates Status (Normal/Cancelled) and Taxation.
                    # Logic needed:
                    # If line[31] == 'C': status='C', tributacao=? (Maybe previous valid one or None?)
                    # If line[31] != 'C': status='N', tributacao=line[31]
                    #
                    # Re-reading Manual 2.3.1:
                    # "Situação do RPS ... C – Cancelado"
                    # "T – Tributado em São Paulo"
                    #
                    # So line[31] (pos 32) holds the Taxation info OR Cancelled status.
                    #
                    # I will implement logic to split this.
                    #
                    valor_servicos_centavos=valor_servicos_centavos,
                    valor_deducoes_centavos=valor_deducoes_centavos,
                    codigo_servico=codigo_servico,
                    aliquota_servicos=aliquota_servicos,
                    iss_retido=iss_retido,
                    serie_rps=serie_rps,
                    tomador=tomador_cpf_cnpj,
                    razao_social_tomador=razao_social_tomador,
                    endereco_tipo_logradouro=endereco_tipo_logradouro,
                    endereco_logradouro=endereco_logradouro,
                    endereco_numero=endereco_numero,
                    endereco_complemento=endereco_complemento,
                    endereco_bairro=endereco_bairro,
                    endereco_uf=endereco_uf,
                    endereco_cep=endereco_cep,
                    email_tomador=email_tomador,
                    discriminacao=discriminacao,
                    valor_pis_centavos=valor_pis_centavos,
                    valor_cofins_centavos=valor_cofins_centavos,
                    valor_inss_centavos=valor_inss_centavos,
                    valor_ir_centavos=valor_ir_centavos,
                    valor_csll_centavos=valor_csll_centavos,
                    valor_carga_tributaria_centavos=valor_carga_tributaria_centavos,
                    percentual_carga_tributaria=percentual_carga_tributaria,
                    fonte_carga_tributaria=fonte_carga_tributaria,
                    codigo_cei=int(codigo_cei) if codigo_cei else None,
                    matricula_obra=int(matricula_obra) if matricula_obra else None,
                    municipio_prestacao=int(municipio_prestacao)
                    if municipio_prestacao
                    else None,
                    numero_encapsulamento=int(numero_encapsulamento)
                    if numero_encapsulamento
                    else None,
                    valor_total_recebido_centavos=int(valor_total_recebido_centavos)
                    if valor_total_recebido_centavos
                    else None,
                    inscricao_municipal_tomador=im_tomador
                    if im_tomador and im_tomador != "00000000"
                    else None,
                    inscricao_estadual_tomador=ie_tomador
                    if ie_tomador and ie_tomador != "000000000000"
                    else None,
                )
                
                # Correction for status/tributacao logic inside the object construction above
                if line[31] == 'C':
                    rps.status_rps = 'C'
                    # Tributacao is mandatory in RPS dataclass, but meaningless if cancelled?
                    # We'll default to 'T' if cancelled, or leave it if it allows.
                    # The dataclass definition says `tributacao_rps: Literal[...]`
                    rps.tributacao_rps = 'T' 
                else:
                    rps.status_rps = 'N'
                    rps.tributacao_rps = line[31]

                lista_rps.append(rps)

        return lista_rps

    def assinatura(self, assinador: Assinador) -> str:
        template = ""
        template += str(self.inscricao_prestador).zfill(8)
        template += self.serie_rps.upper() + (5 - len(self.serie_rps)) * " "
        template += str(self.numero_rps).zfill(12)
        template += self.data_emissao.strftime("%Y%m%d").upper()
        template += self.tributacao_rps
        template += self.status_rps
        template += self.iss_retido == "true" and "S" or "N"
        template += str(self.valor_servicos_centavos).zfill(15)
        template += str(self.valor_deducoes_centavos).zfill(15)
        template += self.codigo_servico.zfill(5)
        if self.tomador_tipo == "CPF":
            template += "1"
        elif self.tomador_tipo == "CNPJ":
            template += "2"
        elif self.tomador_tipo is None:
            template += "3"
        template += (
            (self.tomador or "")
            .replace(".", "")
            .replace("-", "")
            .replace("/", "")
            .zfill(14)
        )

        if self.intermediario is not None:
            if self.intermediario_tipo == "CPF":
                template += "1"
            elif self.intermediario_tipo == "CNPJ":
                template += "2"
            template += (
                (self.intermediario or "")
                .replace(".", "")
                .replace("-", "")
                .replace("/", "")
                .zfill(14)
            )

            template += self.iss_retido_intermediario == "true" and "S" or "N"

        template_bytes = template.encode("ascii")

        signed_template = assinador.assinar_bytes_rsa_sh1(template_bytes)
        return base64.b64encode(signed_template).decode("ascii")

    @property
    def tomador_tipo(self) -> Optional[Literal["CPF", "CNPJ"]]:
        return cpf_ou_cnpj(self.tomador) if self.tomador is not None else None

    @property
    def intermediario_tipo(self) -> Optional[Literal["CPF", "CNPJ"]]:
        return (
            cpf_ou_cnpj(self.intermediario) if self.intermediario is not None else None
        )


@dataclass
class EnvioRPS(RPS, Pedido, Remessa):
    def gerar_xml(self, assinador: Assinador) -> Element:
        xml = self.template.render(
            remetente=self.remetente,
            remetente_tipo=self.remetente_tipo,
            rps=self.gerar_string_xml(assinador),
        )

        return fromstring(xml)

    @property
    def nome_metodo(self):
        return "EnvioRPS"


@dataclass
class RetornoEnvioRpsLote(Retorno):
    numero_lote: int
    inscricao_prestador: int
    remetente: str
    data_envio_lote: date
    qtd_notas_processadas: int
    tempo_processamento: int
    valor_total_servicos: int
    chaves_nfe_rps: List[ChaveNFeRPS]

    @staticmethod
    def ler_xml(xml: Element) -> "RetornoEnvioRps":
        if xml.find(".//Sucesso").text == "false":
            raise ErroEnvioRps(
                codigo=xml.find(".//Codigo").text,
                descricao=xml.find(".//Descricao").text,
            )
        return RetornoEnvioRpsLote(
            numero_lote=int(xml.find(".//NumeroLote").text),
            inscricao_prestador=int(xml.find(".//InscricaoPrestador").text),
            remetente=xml.find(".//CPFCNPJRemetente").find(".//CNPJ").text,
            data_envio_lote=parse(xml.find(".//DataEnvioLote").text).date(),
            qtd_notas_processadas=int(xml.find(".//QtdNotasProcessadas").text),
            tempo_processamento=int(xml.find(".//TempoProcessamento").text),
            valor_total_servicos=int(xml.find(".//ValorTotalServicos").text),
            chaves_nfe_rps=[
                ChaveNFeRPS.ler_xml(retorno)
                for retorno in xml.findall(".//ChaveNFeRPS")
            ],
        )


@dataclass
class EnvioLoteRPS(Pedido, Remessa):
    transacao: bool
    data_inicio_periodo_transmitido: date
    data_fim_periodo_transmitido: date
    lista_rps: List[RPS]

    teste: bool = False
    """
    Define se o lote é um teste ou produção.
    """

    def __post_init__(self):
        if isinstance(self.data_fim_periodo_transmitido, str):
            self.data_fim_periodo_transmitido = parse(
                self.data_fim_periodo_transmitido
            ).date()

        if isinstance(self.data_inicio_periodo_transmitido, str):
            self.data_inicio_periodo_transmitido = parse(
                self.data_inicio_periodo_transmitido
            ).date()

        for idx, rps in enumerate(self.lista_rps):
            if isinstance(rps, dict):
                self.lista_rps[idx] = RPS(**rps)

        assert len(self.lista_rps) > 0, "Deve haver pelo menos um RPS no lote"
        assert len(self.lista_rps) <= 50, "O lote não pode ter mais de 50 RPS"

    def gerar_xml(self, assinador: Assinador) -> Element:
        xml = self.template.render(
            remetente=self.remetente,
            remetente_tipo=self.remetente_tipo,
            transacao=str(self.transacao).lower(),
            dt_inicio=str(self.data_inicio_periodo_transmitido),
            dt_fim=str(self.data_fim_periodo_transmitido),
            qtd_rps=self.quantidade_rps,
            valor_total_servicos=f"{self.valor_total_servicos:.2f}",
            valor_total_deducoes=f"{self.valor_total_deducoes:.2f}",
            lista_rps=[rps.gerar_string_xml(assinador) for rps in self.lista_rps],
        )

        return fromstring(xml)

    @property
    def quantidade_rps(self):
        return len(self.lista_rps)

    @property
    def valor_total_servicos(self):
        return sum(rps.valor_servicos_centavos for rps in self.lista_rps) / 100

    @property
    def valor_total_deducoes(self):
        return sum(rps.valor_deducoes_centavos for rps in self.lista_rps) / 100

    @property
    def remetente_tipo(self) -> Literal["CPF", "CNPJ"]:
        return cpf_ou_cnpj(self.remetente)

    @property
    def metodo(self) -> str:
        if self.teste:
            return "TesteEnvioLoteRPS"
        else:
            return "EnvioLoteRPS"



from ....assinatura import Assinador, AssinadorMock
from zeep.plugins import HistoryPlugin
from zeep import Client, Transport, Settings
from requests import Session
from .pedido import Pedido
from lxml.etree import tostring, fromstring, XMLSchema, ElementBase
from pathlib import Path
from tempfile import mktemp
from .envio_rps import EnvioRPS, RetornoEnvioRps, EnvioLoteRPS, RetornoEnvioRpsLote
from .consulta_cnpj import ConsultaCNPJ, RetornoConsultaCNPJ
from .cancelamento_nfe import CancelamentoNFe, RetornoCancelamentoNFe
from .consulta import ConsultaNFe, RetornoConsulta, ConsultaNFePeriodo


class Cliente:
    assinador: Assinador

    def __init__(self, caminho_pfx: Path, senha_pfx: str):
        self.assinador = Assinador(caminho_pfx, senha_pfx)

    def executar(self, pedido: Pedido) -> ElementBase:
        try:
            history = HistoryPlugin()
            keyfile = Path(mktemp())
            keyfile.write_bytes(self.assinador.private_key_pem_bytes)
            certfile = Path(mktemp())
            certfile.write_bytes(self.assinador.cert_pem_bytes)
            url = "https://nfe.prefeitura.sp.gov.br/ws/lotenfe.asmx?WSDL"
            xml = pedido.gerar_xml(self.assinador)
            session = Session()
            session.cert = (certfile, keyfile)
            settings = Settings(strict=True, xml_huge_tree=True)
            transport = Transport(session=session, cache=None)
            client = Client(
                url, transport=transport, settings=settings, plugins=[history]
            )
            signed_xml = self.assinador.assinar_xml(xml)

            response: str = getattr(client.service, pedido.metodo)(
                1, tostring(signed_xml, encoding=str)
            )

            return fromstring(response.encode("utf-8"))
        finally:
            keyfile.unlink()
            certfile.unlink()

    def gerar_nota(self, pedido: EnvioRPS) -> RetornoEnvioRps:
        return RetornoEnvioRps.ler_xml(self.executar(pedido))

    def gerar_notas_em_lote(self, pedido: EnvioLoteRPS) -> RetornoEnvioRpsLote:
        return RetornoEnvioRpsLote.ler_xml(self.executar(pedido))

    def consultar_cnpj(self, pedido: ConsultaCNPJ) -> RetornoConsultaCNPJ:
        return RetornoConsultaCNPJ.ler_xml(self.executar(pedido))

    def cancelar_nota(self, pedido: CancelamentoNFe) -> RetornoCancelamentoNFe:
        return RetornoCancelamentoNFe.ler_xml(self.executar(pedido))

    def consultar_nota(self, pedido: ConsultaNFe) -> RetornoConsulta:
        return RetornoConsulta.ler_xml(self.executar(pedido))

    def consultar_notas_periodo(self, pedido: ConsultaNFePeriodo) -> RetornoConsulta:
        return RetornoConsulta.ler_xml(self.executar(pedido))


class ClienteMock(Cliente):
    erro: bool

    def __init__(self):
        self.assinador = AssinadorMock()

    def executar(self, pedido: Pedido) -> ElementBase:
        xml = self.assinador.assinar_xml(pedido.gerar_xml(self.assinador))
        pedido_schema_path = (
            Path(__file__).parent
            / "xsds"
            / f"Pedido{pedido.__class__.__name__}_v01.xsd"
        )
        pedido_schema = XMLSchema(file=pedido_schema_path)
        pedido_schema.assertValid(xml)

        retorno_path = (
            Path(__file__).parent / "exemplos" / f"{pedido.classe_retorno}.xml"
        )
        retorno = retorno_path.read_text()
        retorno_xml = fromstring(retorno.encode("utf-8"))

        return retorno_xml



from .pedido import Pedido
from .retorno import Retorno
from dataclasses import dataclass
from typing import Literal
from lxml.etree import Element, fromstring
from abstra_notas.validacoes.cpfcnpj import cpf_ou_cnpj, normalizar_cpf_ou_cnpj
from abstra_notas.assinatura import Assinador
import base64
from .erro import Erro
from .remessa import Remessa


@dataclass
class RetornoCancelamentoNFe(Retorno):
    @staticmethod
    def ler_xml(xml: str):
        sucesso = xml.find(".//Sucesso").text
        if sucesso == "true":
            return RetornoCancelamentoNFe()
        else:
            raise ErroCancelamentoNFe(
                codigo=int(xml.find(".//Codigo").text),
                descricao=xml.find(".//Descricao").text,
            )


@dataclass
class ErroCancelamentoNFe(Erro):
    codigo: int
    descricao: str


@dataclass
class CancelamentoNFe(Pedido, Remessa):
    transacao: bool
    inscricao_prestador: str
    numero_nfe: int

    def __post_init__(self):
        self.remetente = normalizar_cpf_ou_cnpj(self.remetente)

    @property
    def remetente_tipo(self) -> Literal["CPF", "CNPJ"]:
        return cpf_ou_cnpj(self.remetente)

    def gerar_xml(self, assinador: Assinador) -> Element:
        xml = self.template.render(
            remetente=self.remetente,
            remetente_tipo=self.remetente_tipo,
            transacao=str(self.transacao).lower(),
            inscricao_prestador=self.inscricao_prestador,
            numero_nfe=self.numero_nfe,
            assinatura=self.assinatura(assinador),
        ).encode("utf-8")

        return fromstring(xml)

    def assinatura(self, assinador: Assinador) -> str:
        template = ""
        template += self.inscricao_prestador.zfill(8)
        template += str(self.numero_nfe).zfill(12)

        template_bytes = template.encode("ascii")

        signed_template = assinador.assinar_bytes_rsa_sh1(template_bytes)
        return base64.b64encode(signed_template).decode("ascii")
